# standard
from __future__ import (
    annotations,  # used to manage type annotation for method that return Self in Python < 3.11
)

# PyQGIS
from qgis.core import (
    QgsAbstractFeatureIterator,
    QgsCoordinateTransform,
    QgsCsException,
    QgsFeature,
    QgsFeatureRequest,
    QgsGeometry,
)

# plugin
from qduckdb.provider import duckdb_feature_source, duckdb_provider
from qduckdb.toolbelt.log_handler import PlgLogger
from qduckdb.toolbelt.preferences import PlgOptionsManager


class DuckdbFeatureIterator(QgsAbstractFeatureIterator):
    def __init__(
        self,
        source: duckdb_feature_source.DuckdbFeatureSource,
        request: QgsFeatureRequest,
    ):
        """Constructor"""
        # FIXME: Handle QgsFeatureRequest.FilterExpression
        super().__init__(request)
        self._provider: duckdb_provider.DuckdbProvider = source.get_provider()
        self._settings = PlgOptionsManager.get_plg_settings()
        self.log = PlgLogger().log

        self._request = request if request is not None else QgsFeatureRequest()
        self._transform = QgsCoordinateTransform()

        if (
            self._request.destinationCrs().isValid()
            and self._request.destinationCrs() != source._provider.crs()
        ):
            self._transform = QgsCoordinateTransform(
                source._provider.crs(),
                self._request.destinationCrs(),
                self._request.transformContext(),
            )

        try:
            filter_rect = self.filterRectToSourceCrs(self._transform)
        except QgsCsException:
            self.close()
            return

        if not self._provider.isValid():
            return

        geom_column = self._provider.get_geometry_column()

        # Create the list of fields that need to be retrieved
        if self._request.flags() & QgsFeatureRequest.Flag.SubsetOfAttributes:
            list_field_names = [
                self._provider.fields()[idx].name()
                for idx in self._request.subsetOfAttributes()
            ]
        else:
            list_field_names = [field.name() for field in self._provider.fields()]

        fields_name_for_query = ", ".join(list_field_names)
        if fields_name_for_query:
            fields_name_for_query += ","
        self.index_geom_column = len(list_field_names)

        # Create fid/fids list
        feature_id_list = None
        if (
            self._request.filterType() == self._request.FilterFid
            or self._request.filterType() == self._request.FilterFids
        ):
            feature_id_list = (
                [self._request.filterFid()]
                if self._request.filterType() == self._request.FilterFid
                else self._request.filterFids()
            )

        where_clause = ""

        if feature_id_list and not filter_rect.isNull():
            if self._provider.primary_key() == -1:
                where_clause = (
                    f"where st_intersects({geom_column}, "
                    f"st_geomfromtext('{filter_rect.asWktPolygon()}'))"
                    f"and index in {tuple(feature_id_list)}"
                )

            else:
                primary_key_name = list_field_names[self._provider.primary_key()]
                where_clause = (
                    f"where st_intersects({geom_column}, "
                    f"st_geomfromtext('{filter_rect.asWktPolygon()}'))"
                    f"and {primary_key_name} in {tuple(feature_id_list)}"
                )

        if feature_id_list and filter_rect.isNull():
            if self._provider.primary_key() == -1:
                where_clause = f"where index in {tuple(feature_id_list)}"

            else:
                primary_key_name = list_field_names[self._provider.primary_key()]
                where_clause = f"where {primary_key_name} in {tuple(feature_id_list)}"

        if not filter_rect.isNull() and not feature_id_list:
            where_clause = (
                f"where st_intersects({geom_column}, "
                f"st_geomfromtext('{filter_rect.asWktPolygon()}'))"
            )

        if self._request.flags() & QgsFeatureRequest.Flag.NoGeometry:
            base_query = (
                f"select * from (select {fields_name_for_query} "
                f"row_number() over() as index from {self._provider._from_clause})"
            )

        else:
            base_query = (
                f"select * from (select {fields_name_for_query} "
                f"st_aswkb({geom_column}), {geom_column}, row_number() over() as index from {self._provider._from_clause})"
            )

        self.final_query = base_query + f"{where_clause} order by index"

        self._result = self._provider.con().execute(self.final_query)
        self._index = 0

        if self._settings.debug_mode:
            self.log(
                message="feature iterator execute query: {}".format(self.final_query),
                log_level=4,  # 4 = info
                push=False,
            )

    def fetchFeature(self, f: QgsFeature) -> bool:
        """fetch next feature, return true on success

        :param f: Next feature
        :type f: QgsFeature
        :return: True if success
        :rtype: bool
        """
        next_result = self._result.fetchone()

        if not next_result or not self._provider.isValid():
            f.setValid(False)
            return False

        f.setFields(self._provider.fields())
        f.setValid(self._provider.isValid())

        if not self._request.flags() & QgsFeatureRequest.Flag.NoGeometry:
            geometry = QgsGeometry()
            geometry.fromWkb(next_result[self.index_geom_column])
            f.setGeometry(geometry)
            self.geometryToDestinationCrs(f, self._transform)

        if self._provider.primary_key() == -1:
            # the table does not have a primary key, use the row number as fallback
            f.setId(next_result[-1])
        else:
            f.setId(next_result[self._provider.primary_key()])

        # set attributes
        if self._request.flags() & QgsFeatureRequest.Flag.SubsetOfAttributes:
            for idx, attr_idx in enumerate(self._request.subsetOfAttributes()):
                f.setAttribute(attr_idx, next_result[idx])
        else:
            f.setAttributes(list(next_result[: self.index_geom_column]))

        self._index += 1
        return True

    def __iter__(self) -> DuckdbFeatureIterator:
        """Returns self as an iterator object"""
        self._index = 0
        return self

    def __next__(self) -> QgsFeature:
        """Returns the next value till current is lower than high"""
        f = QgsFeature()
        if not self.nextFeature(f):
            raise StopIteration
        else:
            return f

    def rewind(self) -> bool:
        """reset the iterator to the starting position"""
        # virtual bool rewind() = 0;
        if self._index < 0:
            return False
        self._index = 0
        return True

    def close(self) -> bool:
        """end of iterating: free the resources / lock"""
        # virtual bool close() = 0;
        self._index = -1
        return True
